from __future__ import annotations

import traceback
from functools import wraps
from types import MethodType
from types import TracebackType
from typing import Any
from typing import Callable
from typing import cast
from typing import TypeVar

from django.db import connections
from django.db import DEFAULT_DB_ALIAS

from django_perf_rec.operation import AllSourceRecorder
from django_perf_rec.operation import BaseRecorder
from django_perf_rec.operation import Operation
from django_perf_rec.settings import perf_rec_settings
from django_perf_rec.sql import sql_fingerprint


class DBOp(Operation):
    @property
    def name(self) -> str:
        name_parts = ["db"]
        if self.alias != DEFAULT_DB_ALIAS:
            name_parts.append(self.alias)
        return "|".join(name_parts)


LastExecutedQuery = TypeVar("LastExecutedQuery", bound=Callable[..., str])


class DBRecorder(BaseRecorder):
    """
    Monkey-patch-wraps a database connection to call 'callback' on every
    query it runs.
    """

    def __enter__(self) -> None:
        """
        When using the debug cursor wrapper, Django calls
        connection.ops.last_executed_query to get the SQL from the client
        library. Here we wrap this function on the connection to grab the SQL
        as it comes out.
        """
        connection = connections[self.alias]
        self.orig_force_debug_cursor = connection.force_debug_cursor
        connection.force_debug_cursor = True

        def call_callback(func: LastExecutedQuery) -> LastExecutedQuery:
            alias = self.alias
            callback = self.callback

            @wraps(func)
            def inner(self: Any, *args: Any, **kwargs: Any) -> str:
                sql = func(*args, **kwargs)
                hide_columns = perf_rec_settings.HIDE_COLUMNS
                callback(
                    DBOp(
                        alias=alias,
                        query=sql_fingerprint(sql, hide_columns=hide_columns),
                        traceback=traceback.extract_stack(),
                    )
                )
                return sql

            return cast(LastExecutedQuery, inner)

        self.orig_last_executed_query = connection.ops.last_executed_query
        connection.ops.last_executed_query = MethodType(  # type: ignore [assignment]
            call_callback(connection.ops.last_executed_query), connection.ops
        )

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        connection = connections[self.alias]
        connection.force_debug_cursor = self.orig_force_debug_cursor
        connection.ops.last_executed_query = (  # type: ignore [assignment]
            self.orig_last_executed_query
        )


class AllDBRecorder(AllSourceRecorder):
    """
    Launches DBRecorders on all database connections
    """

    sources_setting = "DATABASES"
    recorder_class = DBRecorder



from __future__ import annotations

from traceback import StackSummary
from types import TracebackType
from typing import Any
from typing import Callable

from django.conf import settings

from django_perf_rec.utils import sorted_names


class Operation:
    def __init__(
        self, alias: str, query: str | list[str], traceback: StackSummary
    ) -> None:
        self.alias = alias
        self.query = query
        self.traceback = traceback

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, type(self))
            and self.alias == other.alias
            and self.query == other.query
            and self.traceback == other.traceback
        )

    @property
    def name(self) -> str:
        raise TypeError("Needs implementing in subclass!")


class BaseRecorder:
    def __init__(self, alias: str, callback: Callable[[Operation], None]) -> None:
        self.alias = alias
        self.callback = callback

    def __enter__(self) -> None:
        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        pass


class AllSourceRecorder:
    """
    Launches Recorders on all the active sources
    """

    sources_setting: str
    recorder_class: type[BaseRecorder]

    def __init__(self, callback: Callable[[Operation], None]) -> None:
        self.callback = callback

    def __enter__(self) -> None:
        self.recorders = []
        for name in sorted_names(getattr(settings, self.sources_setting).keys()):
            recorder = self.recorder_class(name, self.callback)
            recorder.__enter__()
            self.recorders.append(recorder)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        for recorder in reversed(self.recorders):
            recorder.__exit__(exc_type, exc_value, exc_traceback)
        self.recorders = []



from __future__ import annotations

from traceback import extract_stack

import pytest
from django.test import SimpleTestCase

from django_perf_rec.operation import Operation


class OperationTests(SimpleTestCase):
    def test_name(self):
        operation = Operation("hi", "world", extract_stack())

        with pytest.raises(TypeError):
            operation.name



from __future__ import annotations

from traceback import extract_stack
from unittest import mock

import pytest
from django.core.cache import caches
from django.test import SimpleTestCase
from django.test import TestCase

from django_perf_rec.cache import AllCacheRecorder
from django_perf_rec.cache import CacheOp
from django_perf_rec.cache import CacheRecorder
from tests.utils import override_extract_stack


class CacheOpTests(SimpleTestCase):
    def test_clean_key_integer(self):
        assert CacheOp.clean_key("foo1") == "foo#"

    def test_clean_key_uuid(self):
        assert CacheOp.clean_key("bdfc9986-d461-4a5e-bf98-8688993abcfb") == "#"

    def test_clean_key_random_hash(self):
        assert CacheOp.clean_key("abc123abc123abc123abc123abc12345") == "#"

    def test_clean_key_session_key_cache_backend(self):
        key = "django.contrib.sessions.cacheabcdefghijklmnopqrstuvwxyz012345"
        assert CacheOp.clean_key(key) == "django.contrib.sessions.cache#"

    def test_clean_key_session_key_cached_db_backend(self):
        key = "django.contrib.sessions.cached_db" + "abcdefghijklmnopqrstuvwxyz012345"
        assert CacheOp.clean_key(key) == "django.contrib.sessions.cached_db#"

    def test_key(self):
        summary = extract_stack()
        op = CacheOp("default", "foo", "bar", summary)
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.query == "bar"
        assert op.traceback == summary

    def test_keys(self):
        op = CacheOp("default", "foo", ["bar", "baz"], extract_stack())
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.query == ["bar", "baz"]

    def test_keys_frozenset(self):
        op = CacheOp("default", "foo", frozenset(["bar", "baz"]), extract_stack())
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.query == ["bar", "baz"]

    def test_keys_dict_keys(self):
        op = CacheOp("default", "foo", {"bar": "baz"}.keys(), extract_stack())
        assert op.alias == "default"
        assert op.operation == "foo"
        assert op.query == ["bar"]

    def test_invalid(self):
        with pytest.raises(ValueError):
            CacheOp("x", "foo", object(), extract_stack())  # type: ignore [arg-type]

    def test_equal(self):
        summary = extract_stack()
        assert CacheOp("x", "foo", "bar", summary) == CacheOp(
            "x", "foo", "bar", summary
        )

    def test_not_equal_alias(self):
        summary = extract_stack()
        assert CacheOp("x", "foo", "bar", summary) != CacheOp(
            "y", "foo", "bar", summary
        )

    def test_not_equal_operation(self):
        summary = extract_stack()
        assert CacheOp("x", "foo", "bar", summary) != CacheOp(
            "x", "bar", "bar", summary
        )

    def test_not_equal_keys(self):
        summary = extract_stack()
        assert CacheOp("x", "foo", ["bar"], summary) != CacheOp(
            "x", "foo", ["baz"], summary
        )

    def test_not_equal_traceback(self):
        assert CacheOp("x", "foo", "bar", extract_stack(limit=1)) != CacheOp(
            "x", "foo", "bar", extract_stack(limit=2)
        )


class CacheRecorderTests(TestCase):
    @override_extract_stack
    def test_default(self, stack_summary):
        callback = mock.Mock()
        with CacheRecorder("default", callback):
            caches["default"].get("foo")
        callback.assert_called_once_with(
            CacheOp("default", "get", "foo", stack_summary)
        )

    @override_extract_stack
    def test_secondary(self, stack_summary):
        callback = mock.Mock()
        with CacheRecorder("second", callback):
            caches["second"].get("foo")
        callback.assert_called_once_with(CacheOp("second", "get", "foo", stack_summary))

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with CacheRecorder("second", callback):
            caches["default"].get("foo")
        assert len(callback.mock_calls) == 0

    def test_record_traceback(self):
        callback = mock.Mock()
        with CacheRecorder("default", callback):
            caches["default"].get("foo")

        assert len(callback.mock_calls) == 1
        assert "django_perf_rec/cache.py" in str(
            callback.call_args_list[0][0][0].traceback
        )


class AllCacheRecorderTests(TestCase):
    @override_extract_stack
    def test_records_all(self, stack_summary):
        callback = mock.Mock()
        with AllCacheRecorder(callback):
            caches["default"].get("foo")
            caches["second"].set("bar", "baz")
            caches["default"].delete_many(["foo"])

        assert callback.mock_calls == [
            mock.call(CacheOp("default", "get", "foo", stack_summary)),
            mock.call(CacheOp("second", "set", "bar", stack_summary)),
            mock.call(CacheOp("default", "delete_many", ["foo"], stack_summary)),
        ]



from __future__ import annotations

try:
    import pytest

    _HAVE_PYTEST = True
except ImportError:
    _HAVE_PYTEST = False

if _HAVE_PYTEST:
    pytest.register_assert_rewrite("django_perf_rec.api")

from django_perf_rec.api import (
    TestCaseMixin,
    get_record_name,
    get_perf_path,
    record,
)  # noqa: E402

__all__ = [
    "TestCaseMixin",
    "get_record_name",
    "get_perf_path",
    "record",
]



from __future__ import annotations

import os
from functools import lru_cache
from threading import local
from types import TracebackType
from typing import Callable

from django_perf_rec import pytest_plugin
from django_perf_rec.cache import AllCacheRecorder
from django_perf_rec.db import AllDBRecorder
from django_perf_rec.operation import Operation
from django_perf_rec.settings import perf_rec_settings
from django_perf_rec.types import PerformanceRecordItem
from django_perf_rec.utils import current_test
from django_perf_rec.utils import record_diff
from django_perf_rec.utils import TestDetails
from django_perf_rec.yaml import KVFile


def get_perf_path(file_path: str) -> str:
    if file_path.endswith(".py"):
        perf_path = file_path[: -len(".py")] + ".perf.yml"
    elif file_path.endswith(".pyc"):
        perf_path = file_path[: -len(".pyc")] + ".perf.yml"
    else:
        perf_path = file_path + ".perf.yml"
    return perf_path


record_current = local()


def get_record_name(
    test_name: str,
    class_name: str | None = None,
    file_name: str = "",
) -> str:
    if class_name:
        record_name = f"{class_name}.{test_name}"
    else:
        record_name = test_name

    # Multiple calls inside the same test should end up suffixing with .2, .3 etc.
    record_spec = (file_name, record_name)
    if getattr(record_current, "record_spec", None) == record_spec:
        record_current.counter += 1
        record_name = record_name + f".{record_current.counter}"
    else:
        record_current.record_spec = record_spec
        record_current.counter = 1

    return record_name


class PerformanceRecorder:
    def __init__(
        self,
        file_name: str,
        record_name: str,
        capture_traceback: Callable[[Operation], bool] | None,
        capture_operation: Callable[[Operation], bool] | None,
    ) -> None:
        self.file_name = file_name
        self.record_name = record_name

        self.record: list[PerformanceRecordItem] = []
        self.db_recorder = AllDBRecorder(self.on_op)
        self.cache_recorder = AllCacheRecorder(self.on_op)
        self.capture_operation = capture_operation
        self.capture_traceback = capture_traceback

    def __enter__(self) -> None:
        self.db_recorder.__enter__()
        self.cache_recorder.__enter__()
        self.load_recordings()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        self.cache_recorder.__exit__(exc_type, exc_value, exc_traceback)
        self.db_recorder.__exit__(exc_type, exc_value, exc_traceback)

        if exc_type is None:
            self.save_or_assert()

    def on_op(self, op: Operation) -> None:
        record = {op.name: op.query}

        if self.capture_operation and not self.capture_operation(op):
            return

        if self.capture_traceback and self.capture_traceback(op):
            record["traceback"] = op.traceback.format()

        self.record.append(record)

    def load_recordings(self) -> None:
        self.records_file = KVFile(self.file_name)

    def save_or_assert(self) -> None:
        orig_record = self.records_file.get(self.record_name, None)
        if perf_rec_settings.MODE == "none":
            assert (
                orig_record is not None
            ), "Original performance record does not exist for {}".format(
                self.record_name
            )

        if orig_record is not None and perf_rec_settings.MODE != "overwrite":
            msg = f"Performance record did not match for {self.record_name}"
            if not pytest_plugin.in_pytest:
                msg += f"\n{record_diff(orig_record, self.record)}"
            assert self.record == orig_record, msg

        self.records_file.set_and_save(self.record_name, self.record)

        if perf_rec_settings.MODE == "all":
            assert (
                orig_record is not None
            ), "Original performance record did not exist for {}".format(
                self.record_name
            )


def record(
    *,
    record_name: str | None = None,
    path: str | None = None,
    capture_traceback: Callable[[Operation], bool] | None = None,
    capture_operation: Callable[[Operation], bool] | None = None,
) -> PerformanceRecorder:
    @lru_cache(maxsize=None)
    def get_test_details() -> TestDetails:
        return current_test()

    if path is None or path.endswith("/"):
        file_name = get_perf_path(get_test_details().file_path)
    else:
        file_name = path

    if path is not None and path.endswith("/"):
        if not os.path.isabs(path):
            directory = os.path.join(
                os.path.dirname(get_test_details().file_path), path
            )
            if not os.path.exists(directory):
                os.makedirs(directory)
        else:
            directory = path

        file_name = os.path.join(directory, os.path.basename(file_name))

    if record_name is None:
        record_name = get_record_name(
            test_name=get_test_details().test_name,
            class_name=get_test_details().class_name,
            file_name=file_name,
        )

    return PerformanceRecorder(
        file_name,
        record_name,
        capture_traceback,
        capture_operation,
    )


class TestCaseMixin:
    """
    Adds record_performance() method to TestCase class it's mixed into
    for easy import-free use.
    """

    def record_performance(
        self, *, record_name: str | None = None, path: str | None = None
    ) -> PerformanceRecorder:
        return record(record_name=record_name, path=path)



from __future__ import annotations

from typing import Dict
from typing import List
from typing import Union

PerformanceRecordItem = Dict[str, Union[str, List[str]]]
PerformanceRecord = List[PerformanceRecordItem]



from __future__ import annotations

from functools import lru_cache
from typing import Container

from sqlparse import parse
from sqlparse import tokens
from sqlparse.sql import Comment
from sqlparse.sql import Comparison
from sqlparse.sql import IdentifierList
from sqlparse.sql import Parenthesis
from sqlparse.sql import Token


@lru_cache(maxsize=500)
def sql_fingerprint(query: str, hide_columns: bool = True) -> str:
    """
    Simplify a query, taking away exact values and fields selected.

    Imperfect but better than super explicit, value-dependent queries.
    """
    parsed_queries = parse(query)

    if not parsed_queries:
        return ""

    parsed_query = sql_recursively_strip(parsed_queries[0])
    sql_recursively_simplify(parsed_query, hide_columns=hide_columns)
    return str(parsed_query).strip()


sql_deleteable_tokens = (
    tokens.Number,
    tokens.Number.Float,
    tokens.Number.Integer,
    tokens.Number.Hexadecimal,
    tokens.String,
    tokens.String.Single,
)


def sql_trim(node: Token, idx: int = 0) -> None:
    tokens = node.tokens
    count = len(tokens)
    min_count = abs(idx)

    while count > min_count and tokens[idx].is_whitespace:
        tokens.pop(idx)
        count -= 1


def sql_strip(node: Token) -> None:
    ws_count = 0

    for token in node.tokens:
        if token.is_whitespace:
            token.value = "" if ws_count > 0 else " "
            ws_count += 1
        else:
            ws_count = 0


def sql_recursively_strip(node: Token) -> Token:
    for sub_node in node.get_sublists():
        sql_recursively_strip(sub_node)

    if isinstance(node, Comment):
        return node

    sql_strip(node)

    # strip duplicate whitespaces between parenthesis
    if isinstance(node, Parenthesis):
        sql_trim(node, 1)
        sql_trim(node, -2)

    return node


def sql_recursively_simplify(node: Token, hide_columns: bool = True) -> None:
    # Erase which fields are being updated in an UPDATE
    if node.tokens[0].value == "UPDATE":
        i_set = [i for (i, t) in enumerate(node.tokens) if t.value == "SET"][0]
        i_wheres = [
            i
            for (i, t) in enumerate(node.tokens)
            if t.is_group and t.tokens[0].value == "WHERE"
        ]
        if i_wheres:
            i_where = i_wheres[0]
            end = node.tokens[i_where:]
        else:
            end = []
        middle = [Token(tokens.Punctuation, " ... ")]
        node.tokens = node.tokens[: i_set + 1] + middle + end

    # Ensure IN clauses with simple value in always simplify to "..."
    if (
        isinstance(node, Comparison)
        and hasattr(node, "tokens")
        and len(node.tokens) == 5
        and node.tokens[2].value.lower() == "in"
        and isinstance(node.tokens[4], Parenthesis)
    ):
        parenthesis = node.tokens[4]
        if all(
            getattr(t, "ttype", "") in sql_deleteable_tokens
            for t in parenthesis.tokens[1:-1]
        ):
            parenthesis.tokens[1:-1] = [Token(tokens.Punctuation, "...")]

    # Erase the names of savepoints since they are non-deteriministic
    if hasattr(node, "tokens"):
        # SAVEPOINT x
        if str(node.tokens[0]) == "SAVEPOINT":
            node.tokens[2].tokens[0].value = "`#`"
            return
        # RELEASE SAVEPOINT x
        elif len(node.tokens) >= 3 and node.tokens[2].value == "SAVEPOINT":
            node.tokens[4].tokens[0].value = "`#`"
            return
        # ROLLBACK TO SAVEPOINT X
        token_values = [getattr(t, "value", "") for t in node.tokens]
        if len(node.tokens) == 7 and token_values[:6] == [
            "ROLLBACK",
            " ",
            "TO",
            " ",
            "SAVEPOINT",
            " ",
        ]:
            node.tokens[6].tokens[0].value = "`#`"
            return

    # Erase volatile part of PG cursor name
    if node.tokens[0].value.startswith('"_django_curs_'):
        node.tokens[0].value = '"_django_curs_#"'

    prev_word_token = None

    for token in node.tokens:
        ttype = getattr(token, "ttype", None)

        # Detect IdentifierList tokens within an ORDER BY, GROUP BY or HAVING
        # clauses
        inside_order_group_having = match_keyword(
            prev_word_token, ["ORDER BY", "GROUP BY", "HAVING"]
        )
        replace_columns = not inside_order_group_having and hide_columns

        if isinstance(token, IdentifierList) and replace_columns:
            token.tokens = [Token(tokens.Punctuation, "...")]
        elif hasattr(token, "tokens"):
            sql_recursively_simplify(token, hide_columns=hide_columns)
        elif ttype in sql_deleteable_tokens:
            token.value = "#"
        elif getattr(token, "value", None) == "NULL":
            token.value = "#"

        if not token.is_whitespace:
            prev_word_token = token


def match_keyword(token: Token | None, keywords: Container[str]) -> bool:
    """
    Checks if the given token represents one of the given keywords
    """
    if not token:
        return False
    if not token.is_keyword:
        return False

    return token.value.upper() in keywords



from __future__ import annotations

from traceback import extract_stack
from traceback import StackSummary
from unittest import mock

from django.test import SimpleTestCase
from django.test import TestCase

from django_perf_rec.db import AllDBRecorder
from django_perf_rec.db import DBOp
from django_perf_rec.db import DBRecorder
from tests.utils import override_extract_stack
from tests.utils import run_query


class DBOpTests(SimpleTestCase):
    def test_create(self):
        op = DBOp("myalias", "SELECT 1", extract_stack())
        assert op.alias == "myalias"
        assert op.query == "SELECT 1"
        assert isinstance(op.traceback, StackSummary)

    def test_equal(self):
        summary = extract_stack()
        assert DBOp("foo", "bar", summary) == DBOp("foo", "bar", summary)

    def test_not_equal_alias(self):
        summary = extract_stack()
        assert DBOp("foo", "bar", summary) != DBOp("baz", "bar", summary)

    def test_not_equal_sql(self):
        summary = extract_stack()
        assert DBOp("foo", "bar", summary) != DBOp("foo", "baz", summary)

    def test_not_equal_traceback(self):
        assert DBOp("foo", "bar", extract_stack(limit=1)) != DBOp(
            "foo", "bar", extract_stack(limit=2)
        )


class DBRecorderTests(TestCase):
    databases = {"default", "second", "replica"}

    @override_extract_stack
    def test_default(self, stack_summary):
        callback = mock.Mock()
        with DBRecorder("default", callback):
            run_query("default", "SELECT 1")
        callback.assert_called_once_with(DBOp("default", "SELECT #", stack_summary))

    @override_extract_stack
    def test_secondary(self, stack_summary):
        callback = mock.Mock()
        with DBRecorder("second", callback):
            run_query("second", "SELECT 1")
        callback.assert_called_once_with(DBOp("second", "SELECT #", stack_summary))

    @override_extract_stack
    def test_replica(self, stack_summary):
        callback = mock.Mock()
        with DBRecorder("replica", callback):
            run_query("replica", "SELECT 1")
        callback.assert_called_once_with(DBOp("replica", "SELECT #", stack_summary))

    def test_secondary_default_not_recorded(self):
        callback = mock.Mock()
        with DBRecorder("second", callback):
            run_query("default", "SELECT 1")
        assert len(callback.mock_calls) == 0

    def test_record_traceback(self):
        callback = mock.Mock()
        with DBRecorder("default", callback):
            run_query("default", "SELECT 1")

        assert len(callback.mock_calls) == 1
        assert "django_perf_rec/db.py" in str(
            callback.call_args_list[0][0][0].traceback
        )


class AllDBRecorderTests(TestCase):
    databases = {"default", "second", "replica"}

    @override_extract_stack
    def test_records_all(self, stack_summary):
        callback = mock.Mock()
        with AllDBRecorder(callback):
            run_query("replica", "SELECT 1")
            run_query("default", "SELECT 2")
            run_query("second", "SELECT 3")

        assert callback.mock_calls == [
            mock.call(DBOp("replica", "SELECT #", stack_summary)),
            mock.call(DBOp("default", "SELECT #", stack_summary)),
            mock.call(DBOp("second", "SELECT #", stack_summary)),
        ]
